import random
import sys
import environment
import optparse
import agent as agent_mod
import gridworld

from gridworldClass import Gridworld


class GridworldEnvironment(environment.Environment):

    def __init__(self, gridWorld):
        self.gridWorld = gridWorld
        self.reset()

    def getCurrentState(self):
        return self.state

    def getPossibleActions(self, state):
        return self.gridWorld.getPossibleActions(state)

    def doAction(self, action):
        successors = self.gridWorld.getTransitionStatesAndProbs(
            self.state, action)
        sum = 0.0
        rand = random.random()
        state = self.getCurrentState()
        for nextState, prob in successors:
            sum += prob
            if sum > 1.0:
                raise ValueError(
                    'Total transition probability more than one; sample failure.')
            if rand < sum:
                reward = self.gridWorld.getReward(state, action, nextState)
                self.state = nextState
                return (nextState, reward)
        raise ValueError(
            'Total transition probability less than one; sample failure.')

    def reset(self):
        self.state = self.gridWorld.getStartState()


def getCliffGrid():
    grid = [[' ', ' ', ' ', ' ', ' '],
            ['S', ' ', ' ', ' ', 10],
            [-100, -100, -100, -100, -100]]
    return Gridworld(grid)


def getCliffGrid2():
    grid = [[' ', ' ', ' ', ' ', ' '],
            [8, 'S', ' ', ' ', 10],
            [-100, -100, -100, -100, -100]]
    return Gridworld(grid)


def getDiscountGrid():
    grid = [[' ', ' ', ' ', ' ', ' '],
            [' ', '#', ' ', ' ', ' '],
            [' ', '#', 1, '#', 10],
            ['S', ' ', ' ', ' ', ' '],
            [-10, -10, -10, -10, -10]]
    return Gridworld(grid)


def getBridgeGrid():
    grid = [['#', -100, -100, -100, -100, -100, '#'],
            [1, 'S',  ' ',  ' ',  ' ',  ' ',  10],
            ['#', -100, -100, -100, -100, -100, '#']]
    return Gridworld(grid)


def getBookGrid():
    grid = [[' ', ' ', ' ', +1],
            [' ', '#', ' ', -1],
            ['S', ' ', ' ', ' ']]
    return Gridworld(grid)


def getMazeGrid():
    grid = [[' ', ' ', ' ', +1],
            ['#', '#', ' ', '#'],
            [' ', '#', ' ', ' '],
            [' ', '#', '#', ' '],
            ['S', ' ', ' ', ' ']]
    return Gridworld(grid)


def getUserAction(state, actionFunction):
    """
    Get an action from the user (rather than the agent).

    Used for debugging and lecture demos.
    """
    import graphicsUtils
    action = None
    while True:
        keys = graphicsUtils.wait_for_keys()
        if 'Up' in keys:
            action = 'north'
        if 'Down' in keys:
            action = 'south'
        if 'Left' in keys:
            action = 'west'
        if 'Right' in keys:
            action = 'east'
        if 'q' in keys:
            sys.exit(0)
        if action == None:
            continue
        break
    actions = actionFunction(state)
    if action not in actions:
        action = actions[0]
    return action


def printString(x):
    print(x)


def runEpisode(agent, environment, discount, decision, display, message, pause, episode):
    returns = 0
    totalDiscount = 1.0
    environment.reset()
    message("BEGINNING EPISODE: "+str(episode)+"\n")
    while True:

        # DISPLAY CURRENT STATE
        state = environment.getCurrentState()
        display(state)
        pause()

        # END IF IN A TERMINAL STATE
        actions = environment.getPossibleActions(state)
        if len(actions) == 0:
            message("EPISODE "+str(episode) +
                    " COMPLETE: RETURN WAS "+str(returns)+"\n")
            return returns

        # GET ACTION (USUALLY FROM AGENT)
        action = decision(state)
        if action == None:
            raise ValueError('Error: Agent returned None action')

        # EXECUTE ACTION
        nextState, reward = environment.doAction(action)
        message("Started in state: "+str(state) +
                "\nTook action: "+str(action) +
                "\nEnded in state: "+str(nextState) +
                "\nGot reward: "+str(reward)+"\n")

        # UPDATE LEARNER
        agent.update(state, action, nextState, reward)
        returns += reward * totalDiscount
        totalDiscount *= discount


def parseOptions():
    optParser = optparse.OptionParser()
    optParser.add_option('-d', '--discount', action='store',
                         type='float', dest='discount', default=0.9,
                         help='Discount on future (default %default)')
    optParser.add_option('-r', '--livingReward', action='store',
                         type='float', dest='livingReward', default=0.0,
                         metavar="R", help='Reward for living for a time step (default %default)')
    optParser.add_option('-n', '--noise', action='store',
                         type='float', dest='noise', default=0.2,
                         metavar="P", help='How often action results in ' +
                         'unintended direction (default %default)')
    optParser.add_option('-e', '--epsilon', action='store',
                         type='float', dest='epsilon', default=0.3,
                         metavar="E", help='Chance of taking a random action in q-learning (default %default)')
    optParser.add_option('-l', '--learningRate', action='store',
                         type='float', dest='learningRate', default=0.5,
                         metavar="P", help='TD learning rate (default %default)')
    optParser.add_option('-i', '--iterations', action='store',
                         type='int', dest='iters', default=10,
                         metavar="K", help='Number of rounds of value iteration (default %default)')
    optParser.add_option('-k', '--episodes', action='store',
                         type='int', dest='episodes', default=0,
                         metavar="K", help='Number of epsiodes of the MDP to run (default %default)')
    optParser.add_option('-g', '--grid', action='store',
                         metavar="G", type='string', dest='grid', default="BookGrid",
                         help='Grid to use (case sensitive; options are BookGrid, BridgeGrid, CliffGrid, MazeGrid, default %default)')
    optParser.add_option('-w', '--windowSize', metavar="X", type='int', dest='gridSize', default=150,
                         help='Request a window width of X pixels *per grid cell* (default %default)')
    optParser.add_option('-a', '--agent', action='store', metavar="A",
                         type='string', dest='agent', default="random",
                         help='Agent type (options are \'random\', \'value\' and \'q\', default %default)')
    optParser.add_option('-t', '--text', action='store_true',
                         dest='textDisplay', default=False,
                         help='Use text-only ASCII display')
    optParser.add_option('-p', '--pause', action='store_true',
                         dest='pause', default=False,
                         help='Pause GUI after each time step when running the MDP')
    optParser.add_option('-q', '--quiet', action='store_true',
                         dest='quiet', default=False,
                         help='Skip display of any learning episodes')
    optParser.add_option('-s', '--speed', action='store', metavar="S", type=float,
                         dest='speed', default=1.0,
                         help='Speed of animation, S > 1.0 is faster, 0.0 < S < 1.0 is slower (default %default)')
    optParser.add_option('-m', '--manual', action='store_true',
                         dest='manual', default=False,
                         help='Manually control agent (for lecture)')

    opts, args = optParser.parse_args()

    # MANAGE CONFLICTS
    if opts.textDisplay or opts.quiet:
        opts.pause = False
        opts.manual = False

    if opts.manual:
        opts.pause = False

    return opts


def run(grid="BookGrid",
        livingReward=0,
        noise=0.2,
        textDisplay=False,
        gridSize=150,
        speed=0.5,
        discount=0.9,
        agent="q",
        episodes=0,
        iters=10,
        learningRate=0.5,
        epsilon=0.3,
        quiet=False,
        pause=False,
        manual=False
        ):

    ###########################
    # GET THE GRIDWORLD
    ###########################

    mdpFunction = getattr(gridworld, "get"+grid)
    mdp = mdpFunction()
    mdp.setLivingReward(livingReward)
    mdp.setNoise(noise)
    env = gridworld.GridworldEnvironment(mdp)

    ###########################
    # GET THE DISPLAY ADAPTER
    ###########################

    import textGridworldDisplay
    display = textGridworldDisplay.TextGridworldDisplay(mdp)
    if not textDisplay:
        import graphicsGridworldDisplay
        display = graphicsGridworldDisplay.GraphicsGridworldDisplay(
            mdp, gridSize, speed)
    display.start()

    ###########################
    # GET THE AGENT
    ###########################

    a = None
    if agent == 'value':
        a = agent_mod.ValueIterationAgent(mdp, discount, iters)
    elif agent == 'q':
        a = agent_mod.QLearningAgent(
            env.getPossibleActions, discount, learningRate, epsilon)
    elif agent == 'sarsa':
        a = agent_mod.SARSA(env.getPossibleActions,
                            discount, learningRate, epsilon)
    elif agent == 'random':
        # No reason to use the random agent without episodes
        if episodes == 0:
            episodes = 1
        a = agent_mod.RandomAgent(mdp.getPossibleActions)
    else:
        raise ValueError('Unknown agent type: ' + agent)

    ###########################
    # RUN EPISODES
    ###########################

    # DISPLAY Q/V VALUES BEFORE SIMULATION OF EPISODES
    if agent == 'value':
        display.displayValues(a, message="VALUES AFTER " +
                              str(iters)+" ITERATIONS")
        display.pause()
        display.displayQValues(
            a, message="Q-VALUES AFTER "+str(iters)+" ITERATIONS")
        display.pause()

    # FIGURE OUT WHAT TO DISPLAY EACH TIME STEP (IF ANYTHING)
    def displayCallback(x): return None
    if not quiet:
        if agent == 'random':
            def displayCallback(state): return display.displayValues(
                a, state, "CURRENT VALUES")
        if agent == 'value':
            def displayCallback(state): return display.displayValues(
                a, state, "CURRENT VALUES")
        if agent == 'q':
            def displayCallback(state): return display.displayQValues(
                a, state, "CURRENT Q-VALUES")
        if agent == 'sarsa':
            def displayCallback(state): return display.displayQValues(
                a, state, "CURRENT Q-VALUES")

    def messageCallback(x): return printString(x)
    if quiet:
        def messageCallback(x): return None

    # FIGURE OUT WHETHER TO WAIT FOR A KEY PRESS AFTER EACH TIME STEP
    def pauseCallback(): return None
    if pause:
        def pauseCallback(): return display.pause()

    # FIGURE OUT WHETHER THE USER WANTS MANUAL CONTROL (FOR DEBUGGING AND DEMOS)
    decisionCallback = a.getAction
    if manual:
        def decisionCallback(state): return getUserAction(
            state, mdp.getPossibleActions)

    # RUN EPISODES
    if episodes > 0:
        print()
        print("RUNNING", episodes, "EPISODES")
        print()
    returns = 0
    for episode in range(1, episodes+1):
        returns += runEpisode(a, env, discount, decisionCallback,
                              displayCallback, messageCallback, pauseCallback, episode)
    if episodes > 0:
        print()
        print("AVERAGE RETURNS FROM START STATE: " +
              str((returns+0.0) / episodes))
        print()
        print()

    # DISPLAY POST-LEARNING VALUES / Q-VALUES
    if agent in ['q', 'sarsa'] and not manual:
        display.displayQValues(
            a, message="Q-VALUES AFTER "+str(episodes)+" EPISODES")
        display.pause()
        display.displayValues(a, message="VALUES AFTER " +
                              str(episodes)+" EPISODES")
        display.pause()


def main():
    opts = parseOptions()
    run(**vars(opts))


if __name__ == '__main__':
    main()
