import numpy as np
from collections import defaultdict
from abc import ABC, abstractmethod

from typing import List, Tuple, Callable


class Agent(ABC):

    @abstractmethod
    def getValue(self, state: Tuple[int, int]) -> float:
        pass

    @abstractmethod
    def getQValue(self, state: Tuple[int, int], action: str) -> float:
        pass

    @abstractmethod
    def getPolicy(self, state: Tuple[int, int]) -> str:
        pass

    @abstractmethod
    def getAction(self, state: Tuple[int, int]) -> str:
        pass

    @abstractmethod
    def update(self, state: Tuple[int, int], action: str, nextState: Tuple[int, int], reward: float) -> None:
        pass

    def reset(self) -> None:
        pass


class QLearningAgent(Agent):

    def __init__(self,
                 actionFunction: Callable[[Tuple[int, int]], List[str]],
                 discount: float = 0.9,
                 learningRate: float = 0.1,
                 epsilon: float = 0.2):
        """
        A Q-Learning agent gets no information about the mdp on
        construction other than a function mapping states to possible actions (actionFunction)
        The other parameters govern its exploration
        strategy and learning rate.
        """
        self.learningRate = learningRate
        self.epsilon = epsilon
        self.discount = discount
        self.actionFunction = actionFunction

        # we initialize the Q_table with a default value of 0
        self.Q_table = defaultdict(lambda: 0.)

    def getValue(self, state: Tuple[int, int]) -> float:
        """Look up the current value of the state.

        Args:
            state (Tuple[int, int]): The state for which the value should be looked up.

        Returns:
            float: The value of the state.
        """
        # TODO return the value of state. Note that we do not have an explicit state value function.
        # What is the State value here if we only have Q-values Available. Remember the state value is the expected return of a state
        # when following the policy.

    def getQValue(self, state: Tuple[int, int], action: str) -> float:
        """Look up the current q-value of the state action pair.

        Args:
            state (Tuple[int, int]): state for which the Q-value should be looked up.
            action (str): Identifier for the action for which the Q-value should be looked up.

        Returns:
            float: The Q-value of the state action pair.
        """
        return self.Q_table[(state, action)]

    def getPolicy(self, state: Tuple[int, int]) -> str:
        """Look up the current action recommendation for the state.

        Args:
            state (Tuple[int, int]): The state for which the action should be looked up.

        Returns:
            str: Identifier of the action that is recommended for the state.
        """
        possible_actions = np.array(self.actionFunction(state))
        if len(possible_actions) > 0:
            # choose action with largest q-value, if multiple actions have the same q-value, choose randomly among them
            q_values = np.array([self.getQValue(state, action)
                                 for action in possible_actions])
            equal_actions = possible_actions[np.argwhere(
                q_values == np.max(q_values))].flatten()
            return np.random.choice(equal_actions)

    def getAction(self, state: Tuple[int, int]) -> str:
        """Choose an action: this will require that your agent balance
          exploration and exploitation as appropriate.

        Args:
            state (Tuple[int, int]): The state for which the action should be chosen.

        Returns:
            str: Identifier of the action that was chosen.
        """
        possible_actions = np.array(self.actionFunction(state))
        # only select an action if there are possible actions, otherwise we are in a terminal state
        if len(possible_actions) > 0:
            # TODO this function should return a action according to an epsilon-greedy strategy
            # getPolicy is probably very handy here
            ...

    def update(self, state: Tuple[int, int], action: str, nextState: Tuple[int, int], reward: float) -> None:
        """Update the Q-table according to the observed transition.

        Args:
            state (Tuple[int, int]): state from with the action was taken
            action (str): the action that was taken
            nextState (Tuple[int, int]): the state that was reached by taking the action
            reward (float): reward that was received for LEAVING state (this is a specific formulation of an MDP used in the Sutton & Barto book)
        """
        # TODO Implement the update according to the Q-learning algorithm
        ...

    def reset(self) -> None:
        """
        called to reset the agent at the beginning of an episode
        """
        # Not needed here
        pass


class SARSA(QLearningAgent):
    # Note that this class is inheriting from QLearningAgent
    def __init__(self, actionFunction, discount=0.9, learningRate=0.1, epsilon=0.2):
        super().__init__(actionFunction, discount, learningRate, epsilon)

        # SARSA needs to know which action it will perform next. That is what this is for.
        self.nextAction = None
        self.nextActionUsed = True

    def getAction(self, state: Tuple[int, int]) -> str:
        """Choose an explorative action. As SARSA is on-policy we need to know what the next action will be during the update step.
        This function will return the action that was chosen during the last update step if it has not been used yet. Otherwise, it will return a new action.

        Args:
            state (Tuple[int, int]): The state for which the action should be chosen.

        Returns:
            str: Identifier of the action that was chosen.
        """
        # Part of the next Action handling
        action = self.nextAction if not self.nextActionUsed else self._getNewAction(
            state)
        self.nextActionUsed = True
        return action

    def _getNewAction(self, state: Tuple[int, int]) -> str:
        """Choose an explorative action. This function will return an action according to an epsilon-greedy strategy.
           Also it sets the nextActionUsed flag to False.

        Args:
            state (Tuple[int, int]): The state for which the action should be chosen.

        Returns:
            str: Identifier of the action that was chosen.
        """
        # Part ot the next action handling
        possible_actions = np.array(self.actionFunction(state))
        if len(possible_actions) > 0:
            self.nextActionUsed = False
            random = np.random.random()
            if random > self.epsilon:
                return self.getPolicy(state)
            else:
                return np.random.choice(possible_actions)

    def update(self,
               state: Tuple[int, int],
               action: str,
               nextState: Tuple[int, int],
               reward: float) -> None:
        """ Update the Q-table according to the observed transition. Also choose the next action to be executed in the environment.

        Args:
            state (Tuple[int, int]): state from with the action was taken
            action (str): action that was taken
            nextState (Tuple[int, int]): state that was reached by taking the action
            reward (float): reward that was received for LEAVING state (this is a specific formulation of an MDP used in the Sutton & Barto book)
        """

        # part of the next action handling
        self.nextAction = self._getNewAction(nextState)

        # TODO implement the update according to the SARSA Algorithm
        ...

    def reset(self) -> None:
        self.nextAction = None
        self.nextActionUsed = True
